#!/usr/bin/perl -w
# ===============================================
# (C) 2017 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

package klinstall;

my $INI_ERROR_TEXT = "";
my $CFG_AUTOANSWERS_FILE = $ENV{"KLAUTOANSWERS"} ? $ENV{"KLAUTOANSWERS"} : "autoanswers.conf";
my $CFG_APPSETUP_FILE = "/var/opt/kaspersky/applications.setup";
my %AUTOANSWERS;

my %PKGMGRS = ("rpm" => "rpm", "deb" => "deb", "pkg" => "pkg", "opkg" => "pkg");
my $IS_FREEBSD = ($^O =~ /^freebsd$/i) ? 1 : 0;

$| = 1;


if ( open FILE, "<$CFG_AUTOANSWERS_FILE" )
{
	foreach (<FILE>)
	{
		s/[\n\r]//g;
		next if !/^(\w+)\s*=\s*(.*)/;
		$AUTOANSWERS{$1} = $2;
	}

	close FILE;
}

1;


sub autoanswer
{
	my $the_key = shift;
    return $AUTOANSWERS{$the_key};
}

sub remove_autoanswer
{
    my $id = shift;

    delete $AUTOANSWERS{$id} if (defined $AUTOANSWERS{$id});
}

sub output
{
	my $text = translate_string(@_) . "\n";
    my $right_margin = 65;
	my $margin = $right_margin - 1;
    my ($para, $accum, $outstr);

    foreach $para (split ("\n", $text))
	{
		$accum = '';
		$para =~ s/\s+/ /g;
		$para .= ' ';

		while ($para =~ s/^([^\n]{0,$margin})\s//)
		{
			$accum .= $1 . "\n";
		}

		$accum .= $para if length ($para) > $margin;
		$outstr .= $accum;
    }

    print $outstr if defined $outstr;
	return 1;
}


sub debug
{
	output("DEBUG: " . translate_string(@_) . "\n");
}

sub fatal
{
	output("\nFatal error: " . translate_string(@_) . "\n");
	exit 2;
}

sub warning
{
	output("Warning: " . translate_string(@_) . "\n");
}

sub translate_string
{
	my ($str, @args) = @_;

	$str =~ s/<B>/\x1B\x5B1m/g;
	$str =~ s/<\/B>/\x1B\x5B22m/g;

	for ( my $i = 0; $i < scalar @args; $i++ )
	{
		warn ("Empty argument $i in string '$str'\n") if !defined $args[$i];
		my $pattern = "%" . ($i + 1);
		$str =~ s/$pattern/$args[$i]/g;
	}

	return $str;
}

sub get_linux_info
{
	my %ld;

	$ld{"NAME"} = "RedHat";
	$ld{"VERSION"} = "7.2";

	return %ld;
}


sub read_app_info
{
	my $id = shift;
	my %app_info;

	open F, "<$CFG_APPSETUP_FILE" or return;
	flock F, 2;
	my $content = join ("", <F>);
	flock F, 8;
	close F;
	$content =~ s/\r//g;

	return unless $content =~ /\[$id\]\n(.*?)\n\n/s;
	$content = $1;

	foreach (split (/\n/, $content))
	{
		next if !/^(.*?)=(.*)$/;

		if ( !defined $app_info{$1} )
		{
			$app_info{$1} = $2;
			next;
		}

		if ( ref $app_info{$1} ne "ARRAY" )
		{
			my @arr = ($app_info{$1}, $2);
			$app_info{$1} = \@arr;
		}
		else
		{
			push @{$app_info{$1}}, $2;
		}
	}

	close F;
	return %app_info;
}


sub write_app_info
{
	my ($appid, %app_info) = @_;
	my $content = "";

	my %olddata = read_app_info ($appid);
	%app_info = (%olddata, %app_info) if (%olddata);

	my $newdata = "[$appid]\n";
	foreach my $key (sort keys(%app_info))
	{
		my $value = defined($app_info{$key}) ? $app_info{$key} : "";

		if ( ref $value eq "ARRAY" )
		{
			foreach my $val (@{$value})
			{
				$newdata .= "$key=$val\n";
			}
			next;
		}

		$newdata .= "$key=$value\n";
	}

	$newdata .= "\n";

	if ( open F, "+<$CFG_APPSETUP_FILE" )
	{
		flock F, 2;
		$content = join ("", <F>);
		$content =~ s/\r//g;
		truncate F, 0;
		seek F, 0, 0;
	}
	else
	{
		open F, ">$CFG_APPSETUP_FILE" or return;
	}

	$content = $content =~ /\[$appid\]\n.*?\n\n/s
			? $` . $newdata . $'
			: "$content$newdata";

	print F $content;
	flock F, 8;
	close F;

	return 1;
}


sub remove_app_info
{
	my $appid = shift;

	open F, "+<$CFG_APPSETUP_FILE" or return;
	flock F, 2;
	my $content = join ("", <F>);
	$content =~ s/\r//g;

	if ( $content =~ /\[$appid\]\n.*?\n\n/s )
	{
		$content = $` . $';
		truncate F, 0;
		seek F, 0, 0;
		print F $content;
	}

	flock F, 8;
	close F;

	unlink $CFG_APPSETUP_FILE if -s $CFG_APPSETUP_FILE == 0;
	return 1;
}

sub internal_dirname
{
	return ($_[0] =~ /(.*)\/[^\/]+$/) ? $1 : $_[0];
}

sub get_os_info
{
	my $arch = `uname -p 2>/dev/null`;
	my $rev = `uname -r`;
	my $name = `uname -s`;
	$arch = `uname -m` if !$arch || $arch =~ /^unknown/;

	$arch =~ s/[\s\r\n]//g;
	$rev =~ s/[\s\r\n]//g;
	$name =~ s/[\s\r\n]//g;

	$rev = $1 if $name =~ /bsd/i && $rev =~ /^([\d\.]+)/;

	return ($arch, $name, $rev);
}


sub display_cmdline_options
{
	my %optlist = @_;
	print "Supported command line options:\n";
	foreach (keys %optlist)
	{
		next if !$optlist{$_};
		print ("    $_ - $optlist{$_}\n");
	}
}

sub check_for_root
{
	my $procname = shift;

	return if $^O eq "cygwin" || $^O eq "MSWin32";
	klinstall::fatal ("%1 must be run by root!", $procname) if $> != 0;
}


sub file_lookup
{
	my ($fn, @paths) = @_;

	foreach my $pathlist (@paths)
	{
		my $path = "$pathlist/$fn";
		$path =~ s/\/\//\//g;
		return $path if -f $path;
	}

	return;
}


sub get_process_pids
{
	my $processname = shift;
	my (%pids, @pidlist);

	my @osdata = get_os_info();

	if ( $osdata[1] =~ /hpux|sunos/i )
	{
		my @plist = split ("\n", `ps -ef`);
		shift @plist;
		foreach (@plist)
		{
			my $pid_part = substr ($_, 9, 6);
			my $cmd_part = substr ($_, 47);
			next if $pid_part !~ /\s*(\d+)\s$/;
			$pid_part = $1;

			$pids{$pid_part} = $cmd_part;
			$pids{$pid_part} = $1 if $pids{$pid_part} =~ /^(.*?)\s+/;
		}
	}
	else
	{
		my @plist = split ("\n", `ps ax`);
		shift @plist;
		foreach (@plist)
		{
			s/\s+$//;
			next if !/^\s*(\d+)/;
			my $pid_part = $1;
			my $cmd_part = substr ($_, 27);
			$pids{$pid_part} = $cmd_part;

			$pids{$pid_part} = $1 if $pids{$pid_part} =~ /^(.*?)\s+/;
		}
	}

	foreach (keys %pids)
	{
		next if $pids{$_} !~ /\b$processname\b/;
		push @pidlist, $_;
	}

	return @pidlist;
}


sub make_path
{
	my ($dirname, $mode, $trackref) = @_;

	return if -d $dirname;
	if ( $dirname =~ /(.+)\/[^\/]+$/ )
	{
		my $msg = make_path ($1, $mode, $trackref);
		return $msg if $msg;
	}

	return "Could not create directory $dirname: $!"
		if !mkdir ($dirname, $mode);

	push (@{$trackref}, $dirname) if $trackref;
	return;
}


sub remove_path
{
	my $dirname = shift;

	return 1 if !-d $dirname;
	remove_path ($1) if $dirname =~ /\/([\/+])$/;

	warning("Could not delete directory $dirname: $!\n") if !rmdir $dirname;
}

sub copy_file
{
	my ($src, $dest) = @_;

	$dest .= "/$1" if -d $dest && $src =~ /([^\/]+)$/;

	open FSRC, "<$src" or return "Could not read file $src: $!";
	open FDST, ">$dest" or return "Could not write file $dest: $!";

	binmode FSRC;
	binmode FDST;
	my $length;

	while ( ($length = sysread (FSRC, my $data, 8192)) > 0 )
	{
		if ( syswrite (FDST, $data, $length) != $length )
		{
			close FSRC;
			close FDST;

			unlink $dest;
			return "Could not write file $dest: $!";
		}
	}

	close FSRC;
	close FDST;

	if ( !defined $length )
	{
		unlink $dest;
		return "Couldn't read file $src: $!";
	}

	my @statinfo = stat ($src);
	if ( @statinfo )
	{
		chown $statinfo[4], $statinfo[5], $dest;
		chmod (($statinfo[2] & 07777), $dest);
	}

	return;
}


sub copy_files
{
	my ($srcdir, $destdir, $pattern) = @_;
	my @errormessages;

	if ( opendir (DIR, $srcdir) )
	{
		my @files = grep { -f "$srcdir/$_" && /$pattern/i } readdir DIR;
		closedir DIR;

		foreach (@files)
		{
			my $msg = klinstall::copy_file ("$srcdir/$_", "$destdir/$_");
			push @errormessages, $msg if $msg;
		}
	}
	else
	{
		push @errormessages, "Could not open directory $srcdir: $!";
	}

	return @errormessages;
}


sub lazy_chown
{
	my $usergroup = shift;
	my @targets = @_;

	my @arr = split (/:/, $usergroup);
	my $user = ($arr[0] && getpwnam($arr[0])) ? getpwnam($arr[0]) : -1;
	my $group = ($arr[1] && getgrnam($arr[1])) ? getgrnam($arr[1]) : -1;

	return if $user == -1 && $group == -1;

	chown ($user, $group, @targets);
}


sub install_file
{
	my ($src, $dest, $mode, $owner) = @_;

	my $msg = copy_file ($src, $dest);
	klinstall::fatal ($msg) if $msg;
	lazy_chown ($owner, $dest);
	chmod (oct($mode), $dest) if $mode;
}


sub install_path
{
	my ($dir, $mode, $owner, $trackref) = @_;

	my $msg = make_path ($dir, oct($mode), $trackref);
	klinstall::fatal ($msg) if $msg;

	lazy_chown ($owner, $dir);
	chmod (oct($mode), $dir);
}


sub ask_question
{
	my ($qid, $def_answer, $question, @args) = @_;

	return $AUTOANSWERS{$qid} if defined $AUTOANSWERS{$qid};

	my $qtext = translate_string ($question, @args);
	chomp ($qtext);

	output ($def_answer ? "$qtext [$def_answer]: " : $qtext);

	my $answer = <STDIN>;
	$answer = "" if !defined $answer;
	$answer =~ s/[\r\n]//g;
	$answer = $def_answer if length($answer) == 0;

	if ( defined $ENV{"KAV_INSTALL_HISTORY_FILE"} && open F, ">>$ENV{KAV_INSTALL_HISTORY_FILE}" )
	{
		print F "$qid=$answer\n";
		close F;
	}

	output("\n");
	return $answer;
}


sub ask_boolean
{
	while ( 1 )
	{
		my $answ = ask_question (@_);

		$answ = "yes" if $answ =~ /^yes$/i || $answ =~ /^y/i;
		$answ = "no"  if $answ =~ /^no$/i  || $answ =~ /^n/i;

		return 1 if $answ eq "yes";
		return 0 if $answ eq "no";

		output("Please answer either 'yes' or 'no'.\n");
	}
}


sub get_work_logic
{
	my ($parsename, $argsref) = @_;
	my ($pkgname, $part) = ("", "");

	$parsename = $1 if $parsename =~ /([^\/]+)$/;

	if ( @{$argsref} && defined $PKGMGRS{@{$argsref}[0]} )
	{
		$pkgname = $PKGMGRS{@{$argsref}[0]};
		shift @{$argsref};
	}

	if ( @{$argsref}
	&& (@{$argsref}[0] eq "pre" || @{$argsref}[0] eq "post") )
	{
		$part = @{$argsref}[0];
		shift @{$argsref};
		return ($part, $pkgname);
	}

	$part = "pre" if $parsename =~ /^pre/;
	$part = "post" if $parsename =~ /^post/;

	return ($part, $pkgname);
}


sub install_AV_bases
{
	my ($srcdir, $destdir) = @_;

	open ASET, "<$srcdir/avp.set" or return;
	my @set_content = <ASET>;
	close ASET;

	foreach (@set_content)
	{
		s/[\n\r]//g;
		next if /^[;#]/ || /^$/;
		return if !-f "$srcdir/$_";
	}

	klinstall::output ("Found Kaspersky Anti-Virus bases in $srcdir, copying them.");
	klinstall::copy_files ($srcdir, $destdir, ".+");

	return 1;
}


sub exec_cmd
{
	my $cmd = shift;
	my $msg;

	$cmd =~ s/([\&;\`'\\\|"*?~<>^\(\)\[\]\{\}\$\n\r])/\\$1/g;

	my $tmpfile = "/tmp/klinstallexec.$$.tmp";
	my $retcode = system ("$cmd 1>$tmpfile 2>&1");
	$retcode = ($retcode == -1 ? "undef" : $?);

	if ( open FILE, "<$tmpfile" )
	{
		$msg = join ("", <FILE>);
		close FILE;

		unlink $tmpfile;
	}

	return ($retcode, $msg);
}


sub exec_cmd_safe
{
	my @args = @_;
	my ($msg, $cmd);

	foreach (@args)
	{
		s/([\&;\`'\\\|"*?~<>^\(\)\[\]\{\}\$\n\r])/\\$1/g;
	}

	my $tmpfile = "/tmp/klinstallexec.$$.tmp";
	my $cmdline = join (" ", @args) . " 1>$tmpfile 2>&1";

	my $retcode = system ($cmdline);
	$retcode = ($retcode == -1 ? "undef" : $?);

	if ( open FILE, "<$tmpfile" )
	{
		$msg = join ("", <FILE>);
		close FILE;

		unlink $tmpfile;
	}

	return ($retcode, $msg);
}


sub exec_zero
{
	my ($rc, $msg) = exec_cmd (@_);

	if ( !defined $rc
	|| ($? >> 8) != 0
	|| ($? & 127) != 0 )
	{
		klinstall::fatal ("execute ($_[0]) failed: $msg");
	}
}


sub get_ini_error
{
	return $INI_ERROR_TEXT ? $INI_ERROR_TEXT : undef;
}

sub load_ini_file
{
	if ( !open FILE, "<$_[0]" )
	{
		$INI_ERROR_TEXT = "unable to read file $_[0]: $!";
		return;
	}

	my $content = join ("", <FILE>);
	close FILE;

	$content =~ s/\r//g;

	my ($sname, $sdata) = ("&%HEADER%&", "");
	my (%inicontent);

	foreach my $line (split ("\n", $content))
	{
		if ( $line =~ /^\s*\[(.*?)\]\s*/ )
		{
			if ( $sname ne "&%HEADER%&" )
			{
				$inicontent{"sections"}{$sname} = $sdata;
				push @{$inicontent{"order"}}, $sname;
			}
			else
			{
				$inicontent{"header"} = $sdata;
			}

			($sname, $sdata) = ($1, "");
			next;
		}

		$sdata .= $line . "\n";
	}

	if ( $sname ne "&%HEADER%&" )
	{
		$inicontent{"sections"}{$sname} = $sdata;
		push @{$inicontent{"order"}}, $sname;
	}
	else
	{
		$inicontent{"header"} = $sdata;
	}

	if ( !$inicontent{"order"} || !@{$inicontent{"order"}} )
	{
		$INI_ERROR_TEXT = "no sections found in file $_[0]";
		return;
	}

	return %inicontent;
}


sub save_ini_file
{
	my $filename = shift;
	my %inidata = @_;

	if ( !open FILE, ">$filename" )
	{
		$INI_ERROR_TEXT = "unable to write file $filename: $!";
		return;
	}

	print FILE "$inidata{header}" if $inidata{"header"};

	foreach my $name (@{$inidata{"order"}})
	{
    	print FILE "[$name]\n$inidata{sections}->{$name}";
    	print FILE "\n" if $inidata{"sections"}->{$name} !~ /\n\n$/;
	}

    close FILE;
	return 1;
}


sub get_ini_value
{
	my ($ref, $section, $key) = @_;

	my $undefpattern = "^$key\\s*=\\s*\$";
	my $pattern = "^$key\\s*=\\s*(.+?)\$";

	return undef if !$ref->{"sections"}->{$section}
		|| $ref->{"sections"}->{$section} =~ /$undefpattern/m
		|| $ref->{"sections"}->{$section} !~ /$pattern/m;

	return $1;
}


sub set_ini_value
{
	my ($ref, $section, %data) = @_;

	if ( !$ref->{"sections"}->{$section} )
	{
		push @{$ref->{"order"}}, $section;
		$ref->{"sections"}->{$section} = "";
	}

	foreach my $key (keys %data)
	{
		my $value = defined $data{$key} ? $data{$key} : "";

		$ref->{"sections"}->{$section} =~ m/^\s*$key\s*=\s*/m
			and $ref->{"sections"}->{$section} =~ s/^\s*$key\s*=.*?\n/$key=$value\n/m
				or $ref->{"sections"}->{$section} = "$key=$value\n$ref->{sections}->{$section}";
	}
}


sub remove_ini_value
{
	my ($ref, $section, @keys) = @_;

	return if !$ref->{"sections"}->{$section};

	foreach my $key (@keys)
	{
		$ref->{"sections"}->{$section} =~ s/^\s*$key\s*=\s*.*\n?//gm;
	}

	if ( !@keys || $ref->{"sections"}->{$section} =~ /^\s*$/s )
	{
		delete $ref->{"sections"}->{$section};
		@{$ref->{"order"}} = grep { $_ ne $section } @{$ref->{"order"}};
	}
}


sub update_ini_file
{
	my ($filename, $section, %data) = @_;

	my %inidata = load_ini_file ($filename);
	return if !%inidata;

	set_ini_value (\%inidata, $section, %data);

	return save_ini_file ($filename, %inidata);
}


sub add_user
{
	my ($username, $group, $shell, $home) = @_;

	return 1 if getpwnam ($username);
	my @cmd = $IS_FREEBSD ? ("pw", "useradd", "-n", $username) : ("useradd");

	push @cmd, ("-g", $group) if $group;
	push @cmd, ("-s", $shell) if $shell;
	push @cmd, ("-d", $home) if $home;

	push @cmd, $username if !$IS_FREEBSD;

	make_path ($home, 0755) if !-d $home;

	return 1 if system (@cmd) == 0;
	$@ = "Could not add user '$username'";
	return undef;
}


sub add_group
{
	my $group = shift;

	return 1 if getgrnam ($group);
	my @cmd = $IS_FREEBSD ? ("pw", "groupadd", "-n", $group) : ("groupadd", $group);

	return 1 if system (@cmd) == 0;
	$@ = "Could not add group '$group'";
	return undef;
}


sub add_user_to_group
{
	my ($user, $group) = @_;

	my $groupmembers = ((getgrnam ($user))[3]);
	my $userrx = quotemeta ($user);
	my @cmd;

	return 1 if $groupmembers && $groupmembers =~ /\b$userrx\b/;

	if ( $IS_FREEBSD ) {
		@cmd = ("pw", "usermod", $user, "-G", $group);
	} elsif ( $^O =~ /^openbsd$/i ) {
		@cmd =("usermod", "-G", $group, $user);
	} else {
		@cmd = ("groupmod", "-A", $user, $group);
	}

	return 1 if system (@cmd) == 0;
	$@ = "Could not add user '$user' to group 'group'";
	return undef;
}

sub exec_cmd_as_user
{
	my ($user, @args) = @_;
	my $uid = getpwnam($user);

	return 255 if !defined $uid;

	my $pid = fork();

	return -1 if $pid == -1;

	if ( !$pid )
	{
		$! = 0;
		$< = $> = $uid;

		exit 255 if $! != 0;

		open (STDOUT, "> /dev/null");
		open (STDERR, "> /dev/null");

		exec @args;
		exit 255;
	}

	wait();
	return $?;
}
